from __future__ import absolute_import
from py_eureka_client.eureka_client import EurekaClient

import os
import time
import posixpath
import asyncio

from log4py import logging
from typing import Dict, List, Any, Optional
from urllib import parse

from mctech_core import get_configure
from mctech_actuator import get_health_manager

from .. import lifecycle
from .eureka_discovery_health_indicator import EurekaDiscoveryCompositeHealthIndicator
from .local_discovery_client import DiscoveryClient

log = logging.getLogger('python.discovery.eurekaDiscoverClient')

event_loop = asyncio.get_event_loop()  # 创建一个默认的事件循环


def _create_client_config(local_address: str,
                          host_name: str, remote_loaded: bool):
    '''
    创建eureka客户端配置信息
    :param local_address
    :param host_name
    :param remote_loaded
    '''
    configure = get_configure()
    info = configure.get_app_info()

    # 准备注册信息
    name = info.get("name")
    port = info.get("port")

    defaultConfig = {
        'perferIpAddress': True,
        'registerEureka': True,
        'client': {
            'serverUrls': 'http://localhost:8761/eureka/'
        }
    }

    config = configure.get_config('eureka', defaultConfig)
    ip_address = config.get("ipAddress", local_address)
    servers = resolve_eureka_servers(config["client"].get("serverUrls"))
    # 启用了远程配置且未加载远程配置时，不注册到eureka上
    register_eureka = False if not remote_loaded else config["registerEureka"]

    management = configure.get_config('management')
    home_page_url = "http://%s:%s" % (ip_address, port)
    prefer_ip_address = True if "preferIpAddress" not in config else config["preferIpAddress"]
    eureka_config = {
        # NOTE: EurekaClient始终都用的delta方式
        # 暂时只在开发和测试环境应用增量更新
        # 'shouldUseDelta': config['shouldUseDelta'],
        'instance': {
            'instanceId': "%s:%s:%d" % (host_name, name, port),
            'hostName': ip_address if prefer_ip_address else host_name,
            'app': name,
            'ipAddr': ip_address,
            'port': port,
            'homePageUrl': home_page_url,
            'statusPageUrl': parse.urljoin(
                home_page_url,
                posixpath.join(management.get("contextPath", "/"), 'info')),
            'healthCheckUrl': parse.urljoin(
                home_page_url,
                posixpath.join(management.get("contextPath", "/"), 'health')),
            'vipAddress': os.environ.get('SERVICE_VIP_ADDRESS', name),
            'metadata': {
                'management.context-path': management.get("contextPath")
            },
            'leaseInfo': {
                'renewalIntervalInSecs': None if not config.get("instance") else config["instance"]["renewalIntervalInSecs"],
                'durationInSecs': None if not config.get("instance") else config["instance"]["durationInSecs"]
            }
        },
        'eureka': {
            'fetchRegistry': True,
            'heartbeatInterval': 5000,
            'registryFetchInterval': 15000,
            'registerWithEureka': register_eureka,
            'preferIpAddress': prefer_ip_address,
            'serviceUrls': servers
        }
    }

    renewal_interval: Optional[int] = config.get("instance", {}).get("renewalIntervalInSecs")
    if renewal_interval and renewal_interval > 0:
        eureka_config['eureka']['heartbeatInterval'] = renewal_interval * 1000

    fetch_interval: Optional[int] = config.get("instance", {}).get("registryFetchIntervalMs")
    if fetch_interval and fetch_interval > 0:
        eureka_config['eureka']['registryFetchInterval'] = fetch_interval
    return eureka_config


def resolve_eureka_servers(server_urls: str):
    '''
    格式化注册中心地址，地址后统一加上/app/
    :param serverUrls 注册中心列表
    '''

    # 在当前版本下不需要处理，使用的eureka包里会自动处理
    return server_urls


def _convert_to_eureka_args(config: dict):
    eureka = config["eureka"]
    instance = config["instance"]
    return {
        'eureka_server': eureka["serviceUrls"],
        'app_name': instance["app"],
        'instance_id': instance["instanceId"],
        'instance_host': instance["hostName"],
        'instance_ip': instance["ipAddr"],
        'instance_port': instance["port"],
        'renewal_interval_in_secs': instance["leaseInfo"]["renewalIntervalInSecs"],
        'home_page_url': instance["homePageUrl"],
        'status_page_url': instance["statusPageUrl"],
        'health_check_url': instance["healthCheckUrl"],
        'vip_adr': instance["vipAddress"],
        'should_register': eureka["registerWithEureka"],
        'metadata': instance["metadata"]
    }


class EurekaDiscoveryClient(DiscoveryClient):
    '''
    服务发现客户端, 注册到发现服务器
    '''

    def __init__(self, local_address: str, host_name: str):
        self._local_address = local_address
        self._host_name = host_name

        # 创建client实例
        self._config = _create_client_config(local_address, host_name, False)
        kwargs = _convert_to_eureka_args(self._config)
        self._client = EurekaClient(**kwargs)

    @property
    def client(self) -> EurekaClient:
        return self._client

    @property
    def local_instance(self):
        return self._config["instance"]

    @property
    def isLocal(self):
        return False

    def start(self):
        '''
        启动客户端
        '''

        self._start(self._client)

    def _start(self, client: Optional[EurekaClient] = None):
        if not client:
            client = self._client

        lifecycle.before_start()
        event_loop.run_until_complete(client.start())
        servers = self._config['eureka']['serviceUrls']
        if client.should_register:
            log.info("发布到注册中心: %s" % servers)
        else:
            log.info("连接到注册中心: %s" % servers)

        oldClient = self._client
        # 切换为新的client
        self._client = client

        if oldClient != client:
            event_loop.run_until_complete(oldClient.stop())

    def register(self):
        # 根据需要把当前应用注册到服务中心
        newConfig = _create_client_config(
            self._local_address, self._host_name, True)
        eureka = newConfig["eureka"]
        should_register_to_eureka = eureka["registerWithEureka"]

        # 重新调用start方法注册到中心
        health_indicator = EurekaDiscoveryCompositeHealthIndicator(self._client, newConfig['instance'], eureka)
        # 注册indicator
        get_health_manager().add_indicator(health_indicator)

        if not should_register_to_eureka:
            # 不需要注册到服务中心
            return

        kwargs = _convert_to_eureka_args(newConfig)
        eureka = EurekaClient(**kwargs)
        self._start(eureka)
        self._config = newConfig
        # 等待2秒
        time.sleep(2)

    def unregister(self):
        event_loop.run_until_complete(self._client.stop())

        # 在注册中心取消当前服务注册后，延迟一个`registryFetchInterval`周期结束
        # 防止服务调用端在没有更新注册信息前仍然把请求发送到当前实例中
        delayMs: int = self._config['eureka']['registryFetchInterval'] + 5000
        time.sleep(delayMs)

    def load_config(self):
        default_config = {
            'id': 'config-server',
            'enabled': True
        }

        configure = get_configure()
        config_server = configure.get_config('config.server', default_config)
        if not config_server["enabled"]:
            return configure

        appInfo = configure.get_app_info()
        name = appInfo.get("name", "")
        active_profiles = configure.active_profiles
        if not active_profiles:
            active_profiles = "default"
        assert isinstance(name, str) and len(name) > 0, "配置文件中未获取到'name'的值"

        remote_path = "/%s/%s/" % (name, active_profiles)

        # 远程服务器上配置文件的路径
        service_id = config_server["id"]
        assert service_id, "配置文件中未获取到'config.server.id'的值"

        # 等待1秒，避免出现未初始化完全的问题
        time.sleep(1)
        future = self.client.do_service(
            config_server["id"],
            remote_path, return_type="json")
        remote_config = event_loop.run_until_complete(future)
        if not remote_config:
            log.info("远程服务器未发现配置文件. file: %s" % remote_path)
            return configure
        log.info("从配置中心获取到配置信息. file: %s" % remote_path)
        assert isinstance(remote_config, Dict)
        property_sources: List[Any] = remote_config["propertySources"]
        # 返回的列表是按profilesy优先级高的在前面，合并的时候需要按优先级低的先合并
        # 所以这里需要取反
        property_sources.reverse()
        for s in property_sources:
            name = s["name"]
            source = s["source"]
            pathname = parse.urlparse(name).path
            filename: str = posixpath.basename(pathname).replace('.yml', '')
            profile = 'default'
            if filename != name and filename != 'appliction':
                lastIndex = filename.rfind('-')
                if lastIndex > 0:
                    profile = filename[lastIndex + 1:]

            params = _resolve_params(source)
            source = _resolve_source(source)
            configure.use(source, name, profile, params)
        return configure


def _resolve_source(source: Dict[str, Any]):
    config = {}
    for key in source.keys():
        c = config
        segments = key.split('.')
        segments.reverse()
        while len(segments) > 1:
            prop_name = segments.pop()
            tmp = c.get(prop_name)
            if tmp is None:
                tmp = {}
                c[prop_name] = tmp
                c = tmp
        val = source[key]
        c[segments[0]] = val
    return config


def _resolve_params(source: Dict[str, Any]):
    params: Dict[str, str] = {}
    keys = list(source.keys())
    for key in keys:
        segments = key.split('.')
        prefix = segments[0]
        if prefix == 'params':
            segments.pop(0)
            param_name = '.'.join(segments)
            params[param_name] = source[key]
            source.pop(key)
    return params



from __future__ import absolute_import
from typing import Union, Optional, Mapping, Tuple, NamedTuple, Any, Literal, Iterable, AsyncIterable
from abc import ABC
from httpx import URL

RequestContent = Union[str, bytes, Iterable[bytes], AsyncIterable[bytes]]
RequestData = Mapping[str, Any]
JsonObject = Any

RpcResponseResult = Union[JsonObject, str, bytes]
RequestContentTypes = Union[JsonObject, RequestContent]

HttpMethod = Literal['get', 'post', 'put', 'delete', 'patch']
MixedMethod = Union[HttpMethod, Literal["ws"]]
ResponseTypes = Literal["json", "bytes", "text", "response"]
Seconds = float

CallMode = Literal["execute", "stream", "ws"]


class HttpRequestBody:
    content: Optional[RequestContent]
    json: Optional[Any]
    # body上的content_type比headers上的更优先
    content_type: Optional[str]

    def __init__(self, *,
                 content: Optional[RequestContent] = None,
                 json: Optional[JsonObject] = None,
                 content_type: Optional[str] = None):
        self.content = content
        self.json = json
        self.content_type = content_type

    def args(self) -> Tuple[Optional[str], RequestContentTypes]:
        if self.content is not None:
            return ("content", self.content)
        elif self.json is not None:
            return ("json", self.json)
        return (None, None)


MixedRequestBody = Union[HttpRequestBody, JsonObject, Mapping[str, Any], str, bytes]


class Invoker(ABC):
    path: str
    params: Optional[Mapping[str, Any]]
    query: Optional[Mapping[str, Any]]
    headers: Optional[Mapping[str, Any]]
    context: Optional[Mapping[str, Any]]

    def __init__(self, path: str, *,
                 params: Optional[Mapping[str, Any]] = None,
                 query: Optional[Mapping[str, Any]] = None,
                 headers: Optional[Mapping[str, Any]] = None,
                 context: Optional[Mapping[str, Any]] = None
                 ):
        '''
        path: api接口端点路径，可以包含':param'这样的参数格式
        params: 当传入的path对应的端点路径中包含参数时，params提供参数的替换值
        query: api接口端点路径的query参数
        headers: 需要传递的额外header头。
                 默认会设置'content-type'为'application/json'，如果需要别的类型可以在此替换
        context: 额外传递的上下文信息，最终会转换成通过header方式传递。
                 与headers参数不一样的是context里的值会把驼峰格式转换成'x-'开头的'-'分隔的小写字符格式。
                 例如tenantId会转换成 'x-tenant-id'，这个参数仅是为了兼容与目前node框架调用的服务
        '''
        self.path = path
        self.params = params
        self.query = query
        self.headers = headers
        self.context = context


class WebSocketInvoker(Invoker):
    pass


class RpcInvoker(Invoker):
    body: Optional[MixedRequestBody]
    response_type: Optional[ResponseTypes]
    retry: Optional[int]
    timeout: Optional[Seconds]

    def __init__(self, path: str, *,
                 response_type: Optional[ResponseTypes] = None,
                 params: Optional[Mapping[str, Any]] = None,
                 query: Optional[Mapping[str, Any]] = None,
                 body: Optional[Union[JsonObject, bytes, str]] = None,
                 headers: Optional[Mapping[str, Any]] = None,
                 retry: Optional[int] = None,
                 context: Optional[Mapping[str, Any]] = None,
                 timeout: Optional[Seconds] = None,
                 json: Optional[bool] = None
                 ):
        '''
        path: api接口端点路径，可以包含':param'这样的参数格式
        params: 当传入的path对应的端点路径中包含参数时，params提供参数的替换值
        query: api接口端点路径的query参数
        body: 传递给调用的api接口的body内容
        headers: 需要传递的额外header头。
                 默认会设置'content-type'为'application/json'，如果需要别的类型可以在此替换
        retry: 出错重试次数，默认值为0。目前并未使用
        context: 额外传递的上下文信息，最终会转换成通过header方式传递。
                 与headers参数不一样的是context里的值会把驼峰格式转换成'x-'开头的'-'分隔的小写字符格式。
                 例如tenantId会转换成 'x-tenant-id'，这个参数仅是为了兼容与目前node框架调用的服务
        timeout: 接口调用客户端超时时间，单位秒。超过设置的超时时间，调用端直接抛出超时异常
        '''
        super().__init__(path, params=params, query=query,
                         headers=headers, context=context)
        self.response_type = response_type
        self.body = body
        # 仅在execute方法内有效，调用返回的类型是否是json,如果为false则返回原始字符串文本
        self.json = True if json is None else json
        self.retry = 0 if retry is None else retry
        self.timeout = timeout


class PipeRpcInvoker(RpcInvoker):
    method: HttpMethod

    def __init__(self, path: str, method: HttpMethod, *,
                 params: Optional[Mapping[str, Any]] = None,
                 query: Optional[Mapping[str, Any]] = None,
                 body: Optional[MixedRequestBody] = None,
                 headers: Optional[Mapping[str, Any]] = None,
                 retry: Optional[int] = None,
                 context: Optional[Mapping[str, Any]] = None,
                 timeout: Optional[Seconds] = None):
        '''
        method: stream调用模式下使用的http请求方法
        path: api接口端点路径，可以包含':param'这样的参数格式
        params: 当传入的path对应的端点路径中包含参数时，params提供参数的替换值
        query: api接口端点路径的query参数
        body: 传递给调用的api接口的body内容
        headers: 需要传递的额外header头。
                 默认会设置'content-type'为'application/json'，如果需要别的类型可以在此替换
        retry: 出错重试次数，默认值为0。目前并未使用
        context: 额外传递的上下文信息，最终会转换成通过header方式传递。
                 与headers参数不一样的是context里的值会把驼峰格式转换成'x-'开头的'-'分隔的小写字符格式。
                 例如tenantId会转换成 'x-tenant-id'，这个参数仅是为了兼容与目前node框架调用的服务
        timeout: 接口调用客户端超时时间，单位秒。超过设置的超时时间，调用端直接抛出超时异常
        '''
        super().__init__(path, params=params, query=query, body=body,
                         headers=headers, retry=retry,
                         context=context, timeout=timeout)
        self.method = method


class RpcServiceInfo:
    reject_unauthorized: bool
    service_id: Optional[str]
    url: Optional[str]
    rpc_type: Optional[str]
    path: Optional[str]
    timeout: Optional[Seconds]

    def __init__(self, *,
                 reject_unauthorized: Optional[bool] = None,
                 service_id: Optional[str] = None,
                 url: Optional[str] = None,
                 rpc_type: Optional[str] = None,
                 path: Optional[str] = None,
                 timeout: Optional[Seconds] = None
                 ):
        '''
        reject_unauthorized: 使用'https'协议时，是否拒绝验证未通过的服务端证书。默认值为true
                             值为true时，表示强制一定要验证证书，且证书必须验证通过
                             值为false时，表示不会验证证书，任何证书都会被当作可信任的对待
        service_id:          需要调用的服务标识（serviceId与url不能同时为空）
        param url:           指定服务的绝对路径的前缀部分。
                             serviceId与url不能同时为空，优先于serviceId
        path:                服务的基准路径，不允许出现参数占位符。
                             设置了此路径时，所有调用该服务的方法都会在调用路径的基础上加上当前路径前缀
        timeout              rpc调用超时时间，默认值为1s，如果RpcInvoker上设置了，以RpcInvoker上的优先
        '''
        self.reject_unauthorized = True if reject_unauthorized is None else reject_unauthorized
        self.service_id = service_id
        self.url = url
        self.path = path
        self.rpc_type = rpc_type
        self.timeout = timeout


class EndPoint(NamedTuple):
    id: str
    host: str
    port: int
    vip_address: bool

    def to_url(self, **kwargs) -> URL:
        kwargs = {**kwargs, 'scheme': "http", 'host': self.host, 'port': self.port}
        return URL(**kwargs)


class RequestOptions(ABC):
    url: Union[URL, str]
    headers: Mapping[str, Any]
    verify: bool
    timeout: Optional[Seconds]

    def __init__(self, url: Union[URL, str], headers: Mapping[str, Any], verify: bool,
                 timeout: Optional[Seconds] = None):
        self.url = url
        self.headers = headers
        self.verify = verify
        self.timeout = timeout


class RpcRequestOptions(RequestOptions):
    method: Optional[HttpMethod]
    body: Optional[MixedRequestBody]
    response_type: ResponseTypes

    def __init__(self, url: Union[URL, str], headers: Mapping[str, Any], verify: bool,
                 method: HttpMethod, response_type: ResponseTypes,
                 body: Optional[MixedRequestBody] = None,
                 timeout: Optional[Seconds] = None):
        super().__init__(url, headers, verify, timeout)
        self.method = method
        self.body = body
        self.response_type = response_type


class WebSocketRequestOptions(RequestOptions):
    pass



from setuptools import setup, find_packages

setup(
    name="mctech_discovery",
    version="1.2.2",
    packages=find_packages(
        include=["mctech_discovery**"],
        exclude=["*.tests"]
    ),
    install_requires=["log4py", "netifaces",
                      "py_eureka_client", "httpx",
                      "mctech-actuator", "mctech-core", "websocket-client"]
)



from __future__ import absolute_import
from typing import Dict, Any
from datetime import datetime
from py_eureka_client.eureka_client import EurekaClient

from mctech_actuator import Health, HealthBuilder, Status, CompositeHealthIndicator


class EurekaDiscoveryCompositeHealthIndicator(CompositeHealthIndicator):
    def __init__(self, client: EurekaClient, instanceConfig: Dict[str, Any], clientConfig: Dict[str, Any]):
        super().__init__()

        discovery = EurekaDiscoveryClientHealthIndicator(client)
        eureka = EurekaHealthIndicator(client, instanceConfig, clientConfig)

        self.add_indicator(discovery._name, discovery)
        self.add_indicator(eureka.name, eureka)

    @property
    def name(self) -> str:
        return 'discoveryComposite'


class EurekaDiscoveryClientHealthIndicator:
    _name: str
    _client: EurekaClient
    _discovery_initialized: bool

    def __init__(self, client: EurekaClient):
        self._name = 'discoveryClient'
        self._client = client
        self._discovery_initialized = False

        # @ts-ignore 本来继承自EventEmitter，但是导出类型描述没有添加
        # client.once('registered', () => {
        #   this._discovery_initialized = true
        # })

    def health(self) -> Health:
        builder = Health.new_builder('UNKNOWN')
        if self._discovery_initialized:
            services = [app.name for app in self._client.applications.applications]
            builder \
                .status('UP', description='Node Cloud Eureka Discovery Client') \
                .add_detail('services', services)
        else:
            builder.status('UNKNOWN', description='Discovery Client not initialized')
        return builder.build()


class EurekaHealthIndicator:
    name: str
    _client: EurekaClient
    _instanceConfig: Dict[str, Any]
    _clientConfig: Dict[str, Any]
    _last_successful_registry_fetch_timestamp: int

    def __init__(self, client: EurekaClient, instanceConfig: Dict[str, Any], clientConfig: Dict[str, Any]):
        self.name = 'eureka'

        self._client = client
        self._instanceConfig = instanceConfig
        self._clientConfig = clientConfig
        self._last_successful_registry_fetch_timestamp = -1
        # @ts-ignore 本来继承自EventEmitter，但是导出类型描述没有添加
        # client.on('heartbeat', () => {
        #   this.last_successful_timestamp = Date.now()
        # })

    def health(self) -> Health:
        builder = Health.unknown()
        status = self._get_status(builder)
        return builder \
            .status(status) \
            .add_detail('applications', self._get_applications()) \
            .build()

    def _get_applications(self):
        services = [app.name for app in self._client.applications.applications]

        apps = {}
        for key in services:
            apps[key] = len(self._client.applications.get_application(key).instances)
        return apps

    def _get_instance_remote_status(self):
        if self._client.should_discover:
            app = self._client.applications.get_application(self._client.__instance['app'])
            if app:
                instance = app.get_instance(self._client.__instance['instanceId'])
                if instance:
                    return instance.status
        return 'UNKNOWN'

    def _get_status(self, builder: HealthBuilder) -> Status:
        status = Status(self._get_instance_remote_status(), 'Remote status from Eureka server')
        if self._clientConfig['fetchRegistry']:
            last_fetch = self._last_successful_registry_fetch_timestamp
            if last_fetch > 0:
                last_fetch = round(datetime.now().timestamp()) - self._last_successful_registry_fetch_timestamp

            if last_fetch < 0:
                status = Status(
                    'UP',
                    'Eureka discovery client has not yet successfully connected to a Eureka server'
                )
            elif last_fetch > self._clientConfig['registryFetchInterval'] * 2000:
                status = Status(
                    'UP',
                    'Eureka discovery client is reporting failures to connect to a Eureka server'
                )

                # 分析源码后可知 __heartbeat_interval 与 __instance['leaseInfo']['renewalIntervalInSecs'] 值完全相同
                builder.add_detail('renewalPeriod', self._instanceConfig['leaseInfo']['renewalIntervalInSecs'] or 30)
                builder.add_detail('failCount', round(last_fetch / self._clientConfig['registryFetchInterval']))

        return status



from __future__ import absolute_import

from typing import Callable, Iterable, AsyncIterable
from websocket import WebSocket

from .ep import EndPointRequest, SimpleEndPointRequest
from .request_context import RequestContext
from ..rpc import lb as lb_factory

from ._rpc_types import WebSocketInvoker, RpcInvoker, PipeRpcInvoker, Invoker, RpcServiceInfo, \
    HttpMethod, MixedMethod, RpcResponseResult, CallMode


RequestCreatorType = Callable[[RequestContext], EndPointRequest]


class AbstractClient:
    _request_creator: RequestCreatorType

    def __init__(self, creator: RequestCreatorType):
        self._request_creator = creator

    async def _do_rpc_request_async(self, method: HttpMethod, invoker: Invoker,
                                    service: RpcServiceInfo, mode: CallMode):
        if mode == 'stream' and not method:
            raise RuntimeError(method, '使用stream方法，invoker.method不能为空')
        r = await self._do_request_async(method, invoker, service, mode)
        assert not isinstance(r, WebSocket)
        return r

    async def _do_request_async(self, method: MixedMethod, invoker: Invoker,
                                service: RpcServiceInfo, mode: CallMode):
        retry = (invoker.retry or 0) if isinstance(invoker, RpcInvoker) else 0 or 0
        lb = lb_factory.create(retry, service)
        context = lb.create_context(method, invoker, service)
        ep_request = self._request_creator(context)

        return await lb.request_async(mode, ep_request)

    def _do_rpc_request(self, method: HttpMethod, invoker: Invoker,
                        service: RpcServiceInfo, mode: CallMode):
        if mode == 'stream' and not method:
            raise RuntimeError(method, '使用stream方法，invoker.method不能为空')
        r = self._do_request(method, invoker, service, mode)
        assert not isinstance(r, WebSocket)
        return r

    def _do_request(self, method: MixedMethod, invoker: Invoker,
                    service: RpcServiceInfo, mode: CallMode):
        retry = (invoker.retry or 0) if isinstance(invoker, RpcInvoker) else 0 or 0
        lb = lb_factory.create(retry, service)
        context = lb.create_context(method, invoker, service)
        ep_request = self._request_creator(context)

        return lb.request(mode, ep_request)


class RpcClient(AbstractClient):
    async def get_async(self, invoker: RpcInvoker, service: RpcServiceInfo) -> RpcResponseResult:
        r = await self._do_rpc_request_async('get', invoker, service, 'execute')
        return r

    async def post_async(self, invoker: RpcInvoker, service: RpcServiceInfo) -> RpcResponseResult:
        r = await self._do_rpc_request_async('post', invoker, service, 'execute')
        return r

    async def put_async(self, invoker: RpcInvoker, service: RpcServiceInfo) -> RpcResponseResult:
        r = await self._do_rpc_request_async('put', invoker, service, 'execute')
        return r

    async def patch_async(self, invoker: RpcInvoker, service: RpcServiceInfo) -> RpcResponseResult:
        r = await self._do_rpc_request_async('patch', invoker, service, 'execute')
        return r

    async def delete_async(self, invoker: RpcInvoker, service: RpcServiceInfo) -> RpcResponseResult:
        r = await self._do_rpc_request_async('delete', invoker, service, 'execute')
        return r

    async def stream_async(self, invoker: PipeRpcInvoker, service: RpcServiceInfo) -> AsyncIterable[bytes]:
        r = await self._do_rpc_request_async(invoker.method, invoker, service, 'stream')
        assert isinstance(r, AsyncIterable)
        return r

    async def ws_async(self, invoker: WebSocketInvoker, service: RpcServiceInfo) -> WebSocket:
        r = await self._do_request_async('ws', invoker, service, 'ws')
        assert isinstance(r, WebSocket)
        return r

    def get(self, invoker: RpcInvoker, service: RpcServiceInfo) -> RpcResponseResult:
        r = self._do_rpc_request('get', invoker, service, 'execute')
        return r

    def post(self, invoker: RpcInvoker, service: RpcServiceInfo) -> RpcResponseResult:
        r = self._do_rpc_request('post', invoker, service, 'execute')
        return r

    def put(self, invoker: RpcInvoker, service: RpcServiceInfo) -> RpcResponseResult:
        r = self._do_rpc_request('put', invoker, service, 'execute')
        return r

    def patch(self, invoker: RpcInvoker, service: RpcServiceInfo) -> RpcResponseResult:
        r = self._do_rpc_request('patch', invoker, service, 'execute')
        return r

    def delete(self, invoker: RpcInvoker, service: RpcServiceInfo) -> RpcResponseResult:
        r = self._do_rpc_request('delete', invoker, service, 'execute')
        return r

    def stream(self, invoker: PipeRpcInvoker, service: RpcServiceInfo) -> Iterable[bytes]:
        r = self._do_rpc_request(invoker.method, invoker, service, 'stream')
        assert isinstance(r, Iterable)
        return r

    def ws(self, invoker: WebSocketInvoker, service: RpcServiceInfo) -> WebSocket:
        r = self._do_request('ws', invoker, service, 'ws')
        assert isinstance(r, WebSocket)
        return r

    def bind(self, service: RpcServiceInfo):
        '''
        需要调用的服务的信息
        '''
        return ServiceBindedRpcClient(self._request_creator, service)

    def set_creator(self, creator: RequestCreatorType):
        if not creator:
            raise RuntimeError('creator不能为空值')

        self._request_creator = creator


class ServiceBindedRpcClient(AbstractClient):
    _service: RpcServiceInfo

    def __init__(self, creator: RequestCreatorType, service: RpcServiceInfo):
        super().__init__(creator)
        self._service = service

    async def get_async(self, invoker: RpcInvoker) -> RpcResponseResult:
        r = await self._do_rpc_request_async('get', invoker, self._service, 'execute')
        return r

    async def post_async(self, invoker: RpcInvoker) -> RpcResponseResult:
        r = await self._do_rpc_request_async('post', invoker, self._service, 'execute')
        return r

    async def put_async(self, invoker: RpcInvoker) -> RpcResponseResult:
        r = await self._do_rpc_request_async('put', invoker, self._service, 'execute')
        return r

    async def patch_async(self, invoker: RpcInvoker) -> RpcResponseResult:
        r = await self._do_rpc_request_async('patch', invoker, self._service, 'execute')
        return r

    async def delete_async(self, invoker: RpcInvoker) -> RpcResponseResult:
        r = await self._do_rpc_request_async('delete', invoker, self._service, 'execute')
        return r

    async def stream_async(self, invoker: PipeRpcInvoker) -> AsyncIterable[bytes]:
        r = await self._do_rpc_request_async(invoker.method, invoker, self._service, 'stream')
        assert isinstance(r, AsyncIterable)
        return r

    async def ws_async(self, invoker: WebSocketInvoker) -> WebSocket:
        r = await self._do_request_async('ws', invoker, self._service, 'ws')
        assert isinstance(r, WebSocket)
        return r

    def get(self, invoker: RpcInvoker) -> RpcResponseResult:
        r = self._do_rpc_request('get', invoker, self._service, 'execute')
        return r

    def post(self, invoker: RpcInvoker) -> RpcResponseResult:
        r = self._do_rpc_request('post', invoker, self._service, 'execute')
        return r

    def put(self, invoker: RpcInvoker) -> RpcResponseResult:
        r = self._do_rpc_request('put', invoker, self._service, 'execute')
        return r

    def patch(self, invoker: RpcInvoker) -> RpcResponseResult:
        r = self._do_rpc_request('patch', invoker, self._service, 'execute')
        return r

    def delete(self, invoker: RpcInvoker) -> RpcResponseResult:
        r = self._do_rpc_request('delete', invoker, self._service, 'execute')
        return r

    def stream(self, invoker: PipeRpcInvoker) -> Iterable[bytes]:
        r = self._do_rpc_request(invoker.method, invoker, self._service, 'stream')
        assert isinstance(r, Iterable)
        return r

    def ws(self, invoker: WebSocketInvoker) -> WebSocket:
        r = self._do_request('ws', invoker, self._service, 'execute')
        assert isinstance(r, WebSocket)
        return r


_rpc_client = RpcClient(lambda context: SimpleEndPointRequest(context))


def get_rpc_client() -> RpcClient:
    return _rpc_client



from __future__ import absolute_import
import socket
import netifaces
from log4py import logging
from typing import Optional

from mctech_core import get_configure

from .local_discovery_client import LocalDiscoveryClient, DiscoveryClient
from .eureka_discovery_client import EurekaDiscoveryClient

_log = logging.getLogger('python.eureka.discoveryClient')


def _get_ip_address():
    '''
    获取本机ip
    '''

    result = []
    for iface in netifaces.interfaces():
        alias = netifaces.ifaddresses(iface)
        ipv4 = alias.get(netifaces.AF_INET)
        if ipv4 is None:
            continue
        for addr_info in ipv4:
            address = addr_info["addr"]
            if address == '127.0.0.1':
                continue
            result.append(addr_info["addr"])

    if len(result) == 0:
        raise RuntimeError('未找到合适的ipv4地址')

    if len(result) > 1:
        _log.warning('找到多个符合条件的ipv4地址: %s' % result)

    selected_address = result[0]
    _log.info('当前使用的ip地址: ' + selected_address)

    return selected_address


def __create_discovery_client() -> DiscoveryClient:
    ip_address = _get_ip_address()
    host_name = socket.gethostname()

    eureka_config = get_configure().get_config('eureka', {'enabled': True})
    discovery = EurekaDiscoveryClient(ip_address, host_name) if eureka_config["enabled"] \
        else LocalDiscoveryClient(ip_address, host_name)
    return discovery


_discovery_client: Optional[DiscoveryClient] = None


def get_discovery() -> DiscoveryClient:
    global _discovery_client
    if not _discovery_client:
        _discovery_client = __create_discovery_client()
    return _discovery_client


__all__ = [
    "get_discovery",
    "DiscoveryClient"
]



from __future__ import absolute_import
import websocket

from datetime import datetime
from log4py import logging
from typing import Iterable, AsyncIterable

from ._lb_types import ServiceLoadBalancer
from ..request_context import InternalRequestContext
from ..ep import EndPointRequest
from ..rule import WeightedResponseTimeRule
from ...discovery import get_discovery

from .._rpc_types import Invoker, RpcServiceInfo, EndPoint, MixedMethod, RpcResponseResult

log = logging.getLogger('python.cloud.rpc.loadbalancer')


class DiscoveryServiceLoadBalancer(ServiceLoadBalancer):
    '''
    通过Eureka服务动态获取可用的服务端地址的实现
    '''

    def __init__(self, service_id: str, app_name: str, product_line: str):
        super().__init__()
        self.service_id = service_id
        self.app_name = app_name
        self.product_line = product_line
        self._rule = WeightedResponseTimeRule(service_id, get_discovery().client)

    def check_alive(self) -> bool:
        # 重新计算各节点权重值
        try:
            if self._rule.check_alive():
                self._rule.compute()
                return True
        except Exception as err:
            log.error('Error calculating server weights', err)

        # 目标服务例不在线或者计算权重失败时都返回false
        return False

    def choose_one(self) -> EndPoint:
        inst = self._rule.choose_server()
        return inst

    def create_context(self,
                       method: MixedMethod,
                       invoker: Invoker,
                       service: RpcServiceInfo) -> InternalRequestContext:
        return InternalRequestContext(method, invoker, service, self.app_name, self.product_line)

    async def execute_async(self, req: EndPointRequest) -> RpcResponseResult:
        async def cb(inst: EndPoint) -> RpcResponseResult:
            result_data = await req.execute_async(inst)
            return result_data
        return await self._execute_with_scope_async(cb)

    async def stream_async(self, req: EndPointRequest) -> AsyncIterable[bytes]:
        async def cb(inst: EndPoint) -> AsyncIterable[bytes]:
            result_stream = await req.stream_async(inst)
            return result_stream
        return await self._execute_with_scope_async(cb)

    async def ws_async(self, req: EndPointRequest) -> websocket.WebSocket:
        async def cb(inst: EndPoint) -> websocket.WebSocket:
            conn = await req.ws_async(inst)
            return conn
        return await self._execute_with_scope_async(cb)

    def execute(self, req: EndPointRequest) -> RpcResponseResult:
        def cb(inst: EndPoint) -> RpcResponseResult:
            result_data = req.execute(inst)
            return result_data
        return self._execute_with_scope(cb)

    def stream(self, req: EndPointRequest) -> Iterable[bytes]:
        def cb(inst: EndPoint) -> Iterable[bytes]:
            result_stream = req.stream(inst)
            return result_stream
        return self._execute_with_scope(cb)

    def ws(self, req: EndPointRequest) -> websocket.WebSocket:
        def cb(inst: EndPoint) -> websocket.WebSocket:
            conn = req.ws(inst)
            return conn
        return self._execute_with_scope(cb)

    async def _execute_with_scope_async(self, callback):
        inst = self._rule.choose_server()

        if log.isEnabledFor(logging.DEBUG):
            log.debug('获取到服务实例: ' + inst.id)
        begin = datetime.now()
        try:
            ret = await callback(inst)
            return ret
        finally:
            end = datetime.now()
            duration = end - begin
            self._rule.record_stats(inst.id, duration.total_seconds())

    def _execute_with_scope(self, callback):
        inst = self._rule.choose_server()
        assert inst is not None, '没有可用的服务实例:' + self.service_id

        if log.isEnabledFor(logging.DEBUG):
            log.debug('获取到服务实例: ' + inst.id)
        begin = datetime.now()
        try:
            ret = callback(inst)
            return ret
        finally:
            end = datetime.now()
            duration = end - begin
            self._rule.record_stats(inst.id, duration.total_seconds())



from __future__ import absolute_import
from typing import Mapping, List, Optional
import json


class WebError(RuntimeError):
    code: Optional[str]
    desc: Optional[str]
    headers: Mapping[str, str]
    status: int

    def __init__(self,
                 message: str,
                 code: Optional[str] = None,
                 desc: Optional[str] = None,
                 headers: Optional[Mapping[str, str]] = None,
                 status: int = 400):
        super().__init__(message)
        self.code = code
        self.desc = desc
        self.headers = headers or {}
        self.status = status
        self.details: List[str] = []

    def __dir__(self):
        # 返回自定义的属性和方法列表
        lst = [k for k in super().__dir__()]
        lst.extend(['code', 'desc', 'headers', 'status', 'details'])
        return lst

    def __str__(self):
        obj = dict(self.__dict__)
        obj['message'] = super().__str__()
        return json.dumps(obj, ensure_ascii=False)



from __future__ import absolute_import
from threading import Lock
from typing import Optional, TypeVar, Generic, Callable
from functools import partial

from httpx._client import BaseClient
from httpx import Client, AsyncClient, Request, Timeout, USE_CLIENT_DEFAULT

from .._rpc_types import Seconds


_CLIENT_RPC_TIMEOUT = Timeout(10.0, connect=3.0)
_ACCEPT_VALUE = 'application/json, application/xml, */*'

_DEFAULT_HEADERS = {
    'Accept': _ACCEPT_VALUE,
    'Accept-Language': 'zh-CN',
    'Accept-Encoding': 'gzip,deflate',
    'User-Agent': 'nodejs rest client',
    # 表示异步调用
    'X-Client-Ajax': 'true'
}

_for_request = BaseClient(headers=_DEFAULT_HEADERS, timeout=_CLIENT_RPC_TIMEOUT)


TClient = TypeVar("TClient")


class _Atomic(Generic[TClient]):
    _lock = Lock()
    _value: Optional[TClient]

    def __init__(self, lazy_create: Callable[[bool], TClient], *kargs):
        self._lazy_create = partial(lazy_create, *kargs)
        self._value = None

    def pop(self) -> Optional[TClient]:
        v = self._value
        if v:
            with self._lock:
                self._value = None
        return v

    def get(self) -> TClient:
        if self._value is None:
            with self._lock:
                if self._value is None:
                    self._value = self._lazy_create()
        return self._value


class _Clients(Generic[TClient]):
    _client: _Atomic[TClient]
    _disable_verify_client: _Atomic[TClient]

    def __init__(self, cls):
        self._client = _Atomic(
            lambda v: cls(headers=_DEFAULT_HEADERS, verify=v, timeout=_CLIENT_RPC_TIMEOUT),
            True
        )
        self._disable_verify_client = _Atomic(
            lambda v: cls(headers=_DEFAULT_HEADERS, verify=v, timeout=_CLIENT_RPC_TIMEOUT),
            False
        )

    def build_request(self, **kwargs) -> Request:
        # 所有的client创建的Request的方法都是一样的，这里随便选择了一个
        return _for_request.build_request(**kwargs)

    def create_timeout(self, timeout: Optional[Seconds]):
        if timeout is None:
            return USE_CLIENT_DEFAULT
        return Timeout(_CLIENT_RPC_TIMEOUT, read=timeout)

    def get_client(self, verify: bool) -> TClient:
        if verify:
            return self._client.get()
        return self._disable_verify_client.get()


class Clients(_Clients[Client]):
    def __init__(self):
        super().__init__(Client)

    def close(self):
        c = self._client.pop()
        if c:
            c.close()
        c = self._disable_verify_client.pop()
        if c:
            c.close()


class AsyncClients(_Clients[AsyncClient]):
    def __init__(self):
        super().__init__(AsyncClient)

    async def aclose(self):
        c = self._client.pop()
        if c:
            await c.aclose()
        c = self._disable_verify_client.pop()
        if c:
            await c.aclose()


clients = Clients()
async_clients = AsyncClients()



from __future__ import absolute_import

from .rpc_client import get_rpc_client, RpcClient
from .request_context import WebError
from .lb import get_service_info

__all__ = [
    "get_rpc_client",
    "get_service_info",
    "RpcClient",
    "WebError"
]



from __future__ import absolute_import
import websocket
from typing import Iterable, AsyncIterable

from ._lb_types import LoadBalancer
from ..request_context import InternalRequestContext
from ..ep import EndPointRequest

from .._rpc_types import Invoker, RpcServiceInfo, EndPoint, MixedMethod, RpcResponseResult


class DirectLoadBalancer(LoadBalancer):
    '''
    显示指定了调用的目标地址的LoadBalancer实现方式
    '''

    def __init__(self):
        super().__init__()

    def check_alive(self):
        return True

    def create_context(self,
                       method: MixedMethod,
                       invoker: Invoker,
                       service: RpcServiceInfo) -> InternalRequestContext:
        return InternalRequestContext(method, invoker, service, None, None)

    @property
    def type(self):
        return 'direct'

    async def execute_async(self, req: EndPointRequest) -> RpcResponseResult:
        ep = self._get_end_point(req)
        result_data = await req.execute_async(ep)
        return result_data

    async def stream_async(self, req: EndPointRequest) -> AsyncIterable[bytes]:
        ep = self._get_end_point(req)
        result_stream = await req.stream_async(ep)
        return result_stream

    async def ws_async(self, req: EndPointRequest) -> websocket.WebSocket:
        ep = self._get_end_point(req)
        conn = await req.ws_async(ep)
        return conn

    def execute(self, req: EndPointRequest) -> RpcResponseResult:
        ep = self._get_end_point(req)
        result_data = req.execute(ep)
        return result_data

    def stream(self, req: EndPointRequest) -> Iterable[bytes]:
        ep = self._get_end_point(req)
        result_stream = req.stream(ep)
        return result_stream

    def ws(self, req: EndPointRequest) -> websocket.WebSocket:
        ep = self._get_end_point(req)
        conn = req.ws(ep)
        return conn

    def _get_end_point(self, req: EndPointRequest) -> EndPoint:
        target_url = req.context.url
        return EndPoint(
            id="%s:%s" % (self.type, req.context.service.url),
            vip_address=False,
            host=target_url.host,
            port=target_url.port or (80 if target_url.scheme == 'http' else 443)
        )



from __future__ import absolute_import
import os
import time
import httpx

from log4py import logging

log = logging.getLogger('python.discovery.lifecycle')


def before_start():
    # 等待sidecar准备好
    _wait_sidecar_ready()


def _wait_sidecar_ready():
    """
    容器中运行时检查sidecar准备好状态

    每秒检查一次， 每5秒输出一条日志
    """

    k8s = os.environ.get('KUBERNETES_SERVICE_HOST')
    sidecar = os.environ.get('ISTIO_INJECT_STATUS')

    if not k8s or not sidecar:
        log.info("There's NO sidecar found. Skip checking sidecar status")
        return

    count = 0
    while True:
        try:
            url = 'http://localhost:15020/healthz/ready'
            res = httpx.get(url, timeout=httpx.Timeout(connect=1, timeout=5))
            res.raise_for_status()
            log.info("The sidecar is ready.")
            break
        except Exception:
            if count % 5 == 0:
                log.info('Waiting for sidecar ready......')
            count = count + 1
            time.sleep(1)



from __future__ import absolute_import
from abc import ABC, abstractmethod
from typing import Any

from mctech_core.config import AppInfo

from mctech_core import get_configure, Configure


class DiscoveryClient(ABC):
    @property
    @abstractmethod
    def isLocal(self) -> bool:
        pass

    @property
    @abstractmethod
    def local_instance(self) -> dict[str, Any]:
        pass

    @property
    @abstractmethod
    def client(self) -> Any:
        pass

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def register(self):
        pass

    @abstractmethod
    def unregister(self):
        pass

    @abstractmethod
    def load_config(self) -> Configure:
        pass


class LocalDiscoveryClient(DiscoveryClient):
    _info: AppInfo
    _local_address: str
    _host_name: str

    def __init__(self, local_address: str, host_name: str):
        self._info = get_configure().get_app_info()
        self._local_address = local_address
        self._host_name = host_name

    @property
    def local_instance(self):
        return {
            'hostName': self._host_name,
            'app': self._info.get("name"),
            'ipAddr': self._local_address,
            'port': self._info.get("port")
        }

    @property
    def isLocal(self):
        return True

    @property
    def client(self):
        raise RuntimeError('LocalDiscoveryClient不支持client属性')

    def start(self):
        # 什么也不做
        pass

    def register(self):
        # 什么也不做
        pass

    def unregister(self):
        # 什么也不做
        pass

    def load_config(self):
        return get_configure()



from __future__ import absolute_import
import unittest
from os import path, environ
environ['NODE_CONFIG_DIR'] = path.join(path.dirname(__file__), "tests")


def _init_discovery():
    from mctech_discovery.discovery import get_discovery
    discovery = get_discovery()
    discovery.start()
    configure = discovery.load_config()
    discovery.register()
    configure.merge()


_init_discovery()
__dir__ = path.dirname(__file__)
discover = unittest.defaultTestLoader.discover(path.join(__dir__, "tests"), pattern="*_test.py")
runner = unittest.TextTestRunner()
runner.run(discover)



from __future__ import absolute_import

import abc
import os
import py_eureka_client.eureka_client as eureka_client
from typing import Optional

from .._rpc_types import EndPoint

k8s = 'KUBERNETES_SERVICE_HOST' in os.environ


class AbstractRule:
    def __init__(self, app_id: str, eureka: eureka_client.EurekaClient):
        self._app_id = app_id
        self._app_id_upper = app_id.upper()
        self._eureka = eureka

    def choose_server(self) -> EndPoint:
        inst = self._choose()
        if not inst:
            raise RuntimeError('没有可用的服务实例: ' + self._app_id)

        if k8s and inst.vipAddress.endswith('.svc'):
            return EndPoint(id=inst.instanceId,
                            host=inst.vipAddress,
                            port=inst.port.port,
                            vip_address=True)

        return EndPoint(id=inst.instanceId,
                        host=inst.ipAddr,
                        port=inst.port.port,
                        vip_address=False)

    @abc.abstractmethod
    def _choose(self) -> Optional[eureka_client.Instance]:
        pass

    def compute(self):
        pass

    def record_stats(self, server_id: str, response_time: float):
        pass



from __future__ import absolute_import
import websocket
from typing import Iterable, AsyncIterable
from abc import ABC, abstractmethod

from ..request_context import RequestContext

from .._rpc_types import EndPoint, RpcResponseResult


class EndPointRequest(ABC):
    def __init__(self, context: RequestContext):
        self.context = context

    @abstractmethod
    async def execute_async(self, ep: EndPoint) -> RpcResponseResult:
        pass

    @abstractmethod
    async def stream_async(self, ep: EndPoint) -> AsyncIterable[bytes]:
        pass

    @abstractmethod
    async def ws_async(self, ep: EndPoint) -> websocket.WebSocket:
        pass

    @abstractmethod
    def execute(self, ep: EndPoint) -> RpcResponseResult:
        pass

    @abstractmethod
    def stream(self, ep: EndPoint) -> Iterable[bytes]:
        pass

    @abstractmethod
    def ws(self, ep: EndPoint) -> websocket.WebSocket:
        pass



import logging

from .discovery import get_discovery
from .rpc import get_rpc_client, get_service_info, RpcClient, WebError
from .rpc._rpc_types import RpcInvoker, PipeRpcInvoker, WebSocketInvoker, RpcServiceInfo

logging.getLogger("httpx").setLevel(logging.WARNING)
logging.getLogger("httpcore").setLevel(logging.INFO)
logging.getLogger("asyncio").setLevel(logging.INFO)


__all__ = [
    "get_discovery",
    "get_rpc_client",
    "get_service_info",
    "RpcClient",
    "WebError",
    "RpcInvoker",
    "PipeRpcInvoker",
    "WebSocketInvoker",
    "RpcServiceInfo"
]
